/**
 * Copyright (c) 2006 Sascha Grau
 * contact: rostar@graui.de
 */

#ifndef DLLIST_H_
#define DLLIST_H_

#include "element.h"

template<class Entry> class DLListIterator;
template<class Entry> class DLListRIterator;

/**
 * A simple double linked list.
 */

template<class Entry> class DLList {

friend class DLListIterator<Entry>;

protected:
    Element<Entry>* first, *last;
    unsigned long elements;

    void ring_insert(Entry e);
    void remove(Element<Entry>* elem);
    virtual void sort(Element<Entry>* l, Element<Entry>* r);
    void exchange(Element<Entry>* l, Element<Entry>* r);

public:
    DLList();
    virtual ~DLList();

    bool empty() const { return elements == 0; }
    unsigned long length() const { return elements; }

    void push_back(Entry e);
    void push_back(DLList<Entry>* list);
    void push_front(Entry e);
    Entry pop_back();
    Entry pop_front();
    void clear();

    void sort();

    Entry front() const;
    Entry back() const;

    Entry at(unsigned long pos) const;

    void print();

    DLListIterator<Entry>* iterator();
    DLListRIterator<Entry>* riterator();
};

/**
 * A double linked list holding pointers.
 * I HATE all this code duplication, but i
 * did not see another way for the needed
 * by-method-template-specalization.
 */
template<class Entry> class DLList<Entry*> {

friend class DLListIterator<Entry*>;

protected:
    Element<Entry*>* first, *last;
    unsigned long elements;

    void ring_insert(Entry* e);
    void remove(Element<Entry*>* elem);
    virtual void sort(Element<Entry*>* l, Element<Entry*>* r);
    void exchange(Element<Entry*>* l, Element<Entry*>* r);

public:

    DLList();
    virtual ~DLList();

    bool empty() const { return elements == 0; }
    unsigned long length() const { return elements; }

    void push_back(Entry* e);
    void push_back(DLList<Entry*>* list);
    void push_front(Entry* e);
    Entry* pop_back();
    Entry* pop_front();
    void clear();

    void sort();

    Entry* front() const;
    Entry* back() const;

    Entry* at(unsigned long pos) const;

    void print();

    DLListIterator<Entry*>* iterator();
    DLListRIterator<Entry*>* riterator();
};


template<class Entry>
class DLListIterator {

private:
    bool first;
    Element<Entry>* pos;
    DLList<Entry>* mylist;

public:
    DLListIterator() {
        pos = NULL;
        mylist = NULL;
        first = true;
    }

    DLListIterator(DLList<Entry>* list) {
        reset(list);
    }

    void reset(DLList<Entry>* list) {
        mylist = list;
        pos = list->first;
        first = true;
    }

    bool hasMore() {
        return ((mylist != NULL) &&
               (pos != mylist->last));
    }
    bool endReached() {
        return (mylist->empty() || ((!first) && (pos == mylist->first)));
    }

    Entry next() { first = false; pos = pos->next; return pos->entry; }
    Entry getElem() { return pos->entry; }
};

template<class Entry>
class DLListRIterator {

private:
    bool first_call;
    Element<Entry>* pos;
    DLList<Entry>* mylist;

public:
    DLListRIterator() {
        pos = NULL;
        mylist = NULL;
        first_call = true;
    }

    DLListRIterator(DLList<Entry>* list) {
        reset(list);
    }

    void reset(DLList<Entry>* list) {
        mylist = list;
        pos = list->last;
        first_call = true;
    }

    bool hasMore() {
        return ((mylist != NULL) &&
               (pos != mylist->first));
    }
    bool beginReached() {
        return (mylist->empty() || ((!first_call) && (pos == mylist->last)));
    }

    Entry prev() { first_call = false; pos = pos->prev; return pos->entry; }
    Entry getElem() { return pos->entry; }
};

/************** inlines **************************************/

template<class Entry>
inline Entry DLList<Entry>::front() const {
    if (front == NULL) return NULL;
    else return static_cast<const Entry>(first->entry);
}

template<class Entry>
inline Entry DLList<Entry>::back() const {
    if (last == NULL) throw -1;
    else return static_cast<const Entry>(last->entry);
}

template<class Entry>
inline void DLList<Entry>::exchange(Element<Entry>* l,
                                    Element<Entry>* r) {
    Entry tmp = l->entry;
    l->entry = r->entry;
    r->entry = tmp;
}

template<class Entry>
inline Entry* DLList<Entry*>::front() const {
    if (front == NULL) return NULL;
    else return static_cast<const Entry*>(first->entry);
}

template<class Entry>
inline Entry* DLList<Entry*>::back() const {
    if (last == NULL) return NULL;
    else return static_cast<const Entry*>(last->entry);
}

template<class Entry>
inline void DLList<Entry*>::exchange(Element<Entry*>* l,
                                    Element<Entry*>* r) {
    Entry* tmp = l->entry;
    l->entry = r->entry;
    r->entry = tmp;
}

#include "dllist.cpp"

#endif /* DLLIST_H_ */
