/**
 * Copyright (c) 2006 Sascha Grau
 * contact: rostar@graui.de
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <map>
#include <iostream>

#include "graphfactory.h"
#include "graphviz/agraph.h"

Graph* GraphFactory::fromFile(char* path)
                    throw(GraphIOException) {

    std::string line;
    Graph* g = new Graph();

    Agraph_t* gviz = NULL;
    //GVC_t* gvc = gvContext();

    FILE* f = fopen(path, "r");

    if (!f) throw GraphIOException("Could not open file " + std::string(path));

    gviz = agread(f, 0);
    fclose(f);

    if (!gviz)
        throw GraphIOException("Could not load graph from file " + std::string(path));

    if (agisdirected(gviz) || !agisstrict(gviz)) {
        throw GraphIOException("Graph has to be undirected & strict.");
    }

    std::map<Agnode_t*, Node*> nodeMap;
    Agnode_t* n;
    Node* mynode;

    /** create an internal representation for each node */
    for (n = agfstnode(gviz); n; n=agnxtnode(n)) {
        mynode = new Node(g->nodes.length(), agnameof(n), agdegree(n, true, true));
        g->nodes.push_back(mynode);
        nodeMap[n] = mynode;
    }

    Agedge_t* e;
    Edge* myedge;

    /** A 'query template' for later 'label'-searches. */
    Agsym_t* attr = agattr(gviz, AGEDGE, "label", 0);

    /** assign edges */
    for (n = agfstnode(gviz); n; n=agnxtnode(n)) {
        for (e = agfstout(n); e; e = agnxtout(e)) {
            char* cost = agxget(e, attr);
            errno = 0;
            long c = strtol(cost, (char**)NULL, 10);
            if ((errno != 0) || (c < 0))
                throw GraphIOException("Graph contains edges with invalid weights: " 
                                                                + std::string(cost));
            Node* a = nodeMap[agtail(e)];
            Node* b = nodeMap[aghead(e)];
            myedge = new Edge(nodeMap[agtail(e)], nodeMap[aghead(e)],
                              static_cast<unsigned long>(c));
            g->edges.push_back(myedge);
            if (!a->addEdge(myedge) || !b->addEdge(myedge)) {
                throw GraphIOException("Edge inconsistencies.");
            }

            /** try to save a little bit of space. */
            agdeledge(e);
            if (a->full())  agdelnode(agtail(e));
            if (b->full())  agdelnode(aghead(e));
        }
    }

    agclose(gviz);

    return g;
}

GraphFactory::GraphFactory() {
    costs.clear();
    permuteBase = NULL;
}

GraphFactory::~GraphFactory() {
    if (permuteBase != NULL)
        delete permuteBase;
}

/** This will NOT copy label information! */
Graph* GraphFactory::copyGraph(Graph* from) {

    std::map<Node*, Node*> nodeMap;
    Graph* g = new Graph();

    DLListIterator<Node*> nit(&from->nodes);
    while (!nit.endReached()) {
        Node* orig = nit.getElem();
        Node* copyNode = new Node();
        copyNode->id = orig->id;
        copyNode->deg_bound = orig->deg_bound;
        nodeMap[orig] = copyNode;
        g->nodes.push_back(copyNode);
        nit.next();
    }

    DLListIterator<Edge*> eit(&from->edges);
    while (!eit.endReached()) {
        Edge* orig = eit.getElem();
        Edge* copyEdge = new Edge();

        Node* tmp = nodeMap[orig->a];
        copyEdge->a = tmp;
        tmp->edges.push_back(copyEdge); 
        tmp = nodeMap[orig->b];
        copyEdge->b = tmp;
        tmp->edges.push_back(copyEdge); 

        copyEdge->cost = orig->cost;
        copyEdge->forest = orig->forest;
        // NO label copy

        g->edges.push_back(copyEdge);
        eit.next();
    }

    return g;
}

Graph* GraphFactory::assignWeights() {

    Graph* g = copyGraph(permuteBase);
    long int index = 0;

    DLListIterator<Edge*> eit(&g->edges);
    while (!eit.endReached()) {
        Edge* e = eit.getElem();
        e->cost = costs[index];

        index++;
        eit.next();
    }

    return g;
}

Graph* GraphFactory::permuteGraphFromFile(char* path)
                throw(GraphIOException) {

    costs.clear();
    if (permuteBase != NULL)
        delete permuteBase;

    permuteBase = fromFile(path);

    costs.resize(permuteBase->edges.length());
    for (unsigned int i=0; i<costs.capacity(); i++) {
        costs[i] = i;
    }

    return assignWeights();
}

 Graph* GraphFactory::permuteGraph() { 
 
    if (permuteBase == NULL) {
        return NULL;
    }

    // all permutations created
    if (!permuteCostVec())
        return NULL;

     return assignWeights(); 

}

bool GraphFactory::permuteCostVec() {

    /** there are better algorithms, but this one
     *  should work. */

    long int k = costs.size();
    long int i,j, tmp;

    i = k - 1;                         // find place to start
    while ((i>0) && (costs[i-1] >= costs[i]))
        i--;

    if (i == 0)
        return false;                  // all in reverse order

    j = k;                           // do next permutation
    while (costs[j-1] <= costs[i-1])
        j--;

    tmp = costs[i-1];
    costs[i-1] = costs[j-1];
    costs[j-1] = tmp;

    i++;
    j = k;

    while (i < j) {                       // need more swaps
        tmp = costs[i-1];
        costs[i-1] = costs[j-1];
        costs[j-1] = tmp;
        i++;
        j--;
    }
    return true;
}












