/**
 * Copyright (c) 2006 Sascha Grau
 * contact: rostar@graui.de
 */

#include "ds/dllist.h"
#include "ds/queue.h"
#include "ds/multipart.h"
#include "io/graphfactory.h"
#include "io/graphexport.h"

#include <errno.h>
#include <iostream>

bool permutation_mode;
bool segment;
bool event;
int verbose;
#define V_NONE 0
#define V_V    1
#define V_VV   2
#define V_VVV  3

Multipart* part;

void dbg_print_q(DLList<Edge*>* q) {

    DLListIterator<Edge*> it;
    it.reset(q);
    while (!it.endReached()) {
        std::cout << " " << it.getElem()->cost;
        it.next();
    }
    std::cout << std::endl;
}

bool inline inClump(Edge* e) {
    return part->find(e->a->id, 0) 
            == part->find(e->b->id, 0);
}

bool inline isInsertable(Edge* e, unsigned int forest) {
    return part->find(e->a->id, forest) 
           != part->find(e->b->id, forest);
}

/** Does the labelling step and returns a pointer
 *  to the last edge of the augmenting path or
 *  NULL, if there is no aufmenting path. */
Edge* label(Graph* g, Edge* e, unsigned int k) {

    int dbg = 0;

    /** Initialization: choose root */
    Node* x = e->a;
    Queue<Node*> q;

    /** reset edge labels and upwards marks */
    DLListIterator<Edge*> eit;
    DLListIterator<Node*>* nit = g->nodes.iterator();
    while(!nit->endReached()) {
        Node* cur = nit->getElem();
        cur->upwards.clear();
        eit.reset(&cur->edges);
        while(!eit.endReached()) {
            eit.getElem()->label = NULL;
            eit.next();
        }
        nit->next();
    }
    delete nit;

    dbg = 1;

    /** root all trees containing x at x. */
    for (unsigned int i=1; i<=k; i++) {
        x->upwards[i] = NULL;
        q.enq(x);
        while (!q.empty()) {
            Node* cur = q.deq();
            eit.reset(&cur->edges);
            while (!eit.endReached()) {
                Edge* out = eit.getElem();
                if ((out->forest == i) && (cur->upwards[i] != out)) {
                    Node* peer = out->traverse(cur);
                    peer->upwards[i] = out;
                    q.enq(peer);
                }
                eit.next();
            }
        }
    }

    dbg = 2;

    /** Here comes the real labelling step. */
    Queue<Edge*> labelq;
    DLList<Edge*> labelstack;
    labelq.enq(e);
    while (!labelq.empty()) {
        Edge* cur = labelq.deq();

        if ((cur != e) && isInsertable(cur, 1 + (cur->forest % k)))
            return cur;

        Node* lower;
        if ((cur->a == x) || (cur->a->upwards[cur->forest] != cur)) {
            lower = cur->b;
        } else {
            lower = cur->a;
        }

        Edge* up = lower->upwards[1 + (cur->forest % k)];
        while ((lower != x) && (up->label == NULL)) {
            up->label = cur;
            labelstack.push_front(up);
            lower = up->traverse(lower);
            up = lower->upwards[1 + (cur->forest % k)];
        }

        // add labels in reverse order to the queue
        labelq.push_back(&labelstack);
    }
    return NULL;
}

void inline swap(Graph* g, Edge* e, unsigned int k) {

    unsigned long swaps = 0;
    Edge* orig = e;

    if (e != NULL)
        part->unite(e->a->id, e->b->id, 1 + (e->forest % k));

    while (e != NULL) {
        swaps++;
        e->forest = 1 + (e->forest % k);
        e = e->label;
    }

    if (verbose >= V_V) {
        if (swaps > k) {
            std::cout << "Swap sequence longer than k." << std::endl;
            std::cout << "Swap sequence length " << swaps;
            std::cout << "." << std::endl;
            while (orig != NULL) {
                std::cout << orig->cost << " ";
                orig = orig->label;
            }
            std::cout << std::endl;
            event = true;
        } else if (verbose == V_VVV) {
            std::cout << "Swap sequence length " << swaps;
            std::cout << "." << std::endl;
            while (orig != NULL) {
                std::cout << orig->cost << " ";
                orig = orig->label;
            }
            std::cout << std::endl;
        }
    }
}

void usage(char* progname) {

    std::cout << "usage:" << std::endl;
    std::cout << progname << " [-o <outfile>] [-k <forestnum>] [-v[v[v]]] <infile>"  << std::endl;
    std::cout << "options:" << std::endl;
    std::cout << "-o\tFile containing the resulting graph. Default: out.dot" << std::endl;
    std::cout << "-k\tNumber of forest. Default: 1 (this is simple MST)" << std::endl;
    std::cout << "\tYou probably want to change this." << std::endl;
    std::cout << "-s\tCreate additional output files containing the forest-segmented graph." << std::endl;
    std::cout << "-v\tReport important events." << std::endl;
    std::cout << "-vv\tAdditionally give insertion operation reports." << std::endl;
    std::cout << "--permute-weights\tIgnore edge weights, instead assign running numbers" << std::endl;
    std::cout << "                 \tand run the algorithm on all possible cost permutations" << std::endl;
    std::cout << "                 \tof graph edges. No file output. Best used with -v." << std::endl;
    std::cout << "                 \tUse with care!" << std::endl;
}

/** process cmd-line arguments */
void process_cmd_line(int argc, char* argv[], 
                      char*& in, std::string& out, 
                      unsigned int& k) {

    int arg = 1;
    bool showusage = (argc < 2);
    while((!showusage) && (arg < argc)) {

        if (strcmp(argv[arg], "-vvv") == 0) {
            verbose = V_VVV;
        } else if (strcmp(argv[arg], "-vv") == 0) {
            verbose = V_VV;
        } else if (strcmp(argv[arg], "-v") == 0) {
            verbose = V_V;
        } else if (strcmp(argv[arg], "--permute-weights") == 0) {
            permutation_mode = true;
        } else if (strcmp(argv[arg], "-s") == 0) {
            segment = true;
        } else if (strcmp(argv[arg], "-o") == 0) {
            if (argc - arg < 2) {
                showusage = true;
            } else {
                arg++;
                if (strstr(argv[arg], "-"))
                    showusage = true;
                else
                    out = std::string(argv[arg]);
            }
        } else if (strcmp(argv[arg], "-k") == 0) {
            if (argc - arg < 2) {
                showusage = true;
            } else {
                arg++;
                if (strstr(argv[arg], "-"))
                    showusage = true;
                else {
                    errno = 0;
                    int tmp = strtol(argv[arg], (char **)NULL, 10);;
                    if ((errno != 0) || (tmp <= 0))
                        showusage = true;
                    else
                        k = static_cast<unsigned int>(tmp);
                }
            }
        } else if (arg == argc-1) {
            in = argv[arg];
        } else {
            showusage = true;
        }

        arg++;
    }

    if (showusage) {
        usage(argv[0]);
        return exit(-1);
    }
}

void rostar(Graph* g, unsigned int k) {

    unsigned long good_edges = 0;

    part = new Multipart(g->nodeCount(), k+1);

    g->edges.sort();

    DLListIterator<Edge*>* it = g->edges.iterator();
    while (!it->endReached() && 
           (good_edges < k*(g->nodes.length() - 1))) {

        Edge* e = it->getElem();
        if (verbose >= V_VV) {
            std::cout << "Trying edge " << e->cost << std::endl;
        }

        if (!inClump(e)) {

            if (isInsertable(e, 1)) {
                e->forest = 1;
                part->unite(e->a->id, e->b->id, 1);
                if (verbose >= V_VV) {
                    std::cout << "Directly insertable." << std::endl;
                }
            } else {
                Edge* last = label(g, e, k);
                if (last == NULL) {
                    // a clump edge
                    // remember the clump nodes...
                    part->unite(e->a->id, e->b->id, 0);
                    if (verbose >= V_VV) {
                        std::cout << "Insertion failed." << std::endl;
                    }
                } else {
                    swap(g, last, k);
                    good_edges++;
                }
            }

            if (verbose >= V_VVV) {
                GraphExport::printGraph(g);
                std::cout << std::endl;
            }

        } else {
            if (verbose >= V_VV) {
                std::cout << "Edge was in clump." << std::endl;
            }
        }

        it->next();
    }
    delete it;

    if (verbose >= V_VV) {
        int trees = 0;
        for (unsigned int i=1; i<=k; i++) {
            if (part->part_cnt(i) == 1)
                trees++;
        }
        std::cout << "Graph contains " << trees << " tree(s) and ";
        std::cout << part->part_cnt(0) << " clumps." << std::endl;
    }

    delete part;
}

int main(int argc, char* argv[]) {

    /** init */
    char* file;
    std::string outfile = "out.dot";
    unsigned int k = 1;
    verbose = V_NONE;
    permutation_mode = false;
    segment = false;
    event = false;
    GraphFactory* gf;

    process_cmd_line(argc, argv, file, outfile, k);

    /** import */
    Graph* g;
    try {
        if (!permutation_mode) {
            g = GraphFactory::fromFile(file);
        } else {
            gf = new GraphFactory();
            g = gf->permuteGraphFromFile(file);
        }
    } catch(GraphIOException exc) {
        std::cout << exc.getCause() << std::endl;
        return -1;
    }

    if (verbose >= V_VV) {
        std::cout << "Graph has " << g->nodes.length() << " nodes and ";
        std::cout << g->edges.length() << " edges." << std::endl;
    }

    /** work */
    long int perm = 0;

    if (!permutation_mode) {
        rostar(g, k);
        if (verbose >= V_VV)
            GraphExport::printGraph(g);
    } else {
        do {
            event = false;
            ++perm;
            rostar(g, k);
            if (verbose >= V_VV) {
                std::cout << "Result:" << std::endl;
                GraphExport::printGraph(g);
            }
            if (event && (verbose == V_V)) {
                std::cout << "Event in permutation " << perm << std::endl;
                GraphExport::toDotStream(std::cout, g);
                return 0;
            }
            delete g;
            g = gf->permuteGraph();
        } while(g != NULL);
    }

    /** export */
    if (!permutation_mode) {
        try {
            GraphExport::toDotFile(outfile, g);
            if (segment) {
                for(unsigned int i=0; i<=k; i++) {
                    char buffer[40];
                    buffer[0] = '.';
                    sprintf(&buffer[1], "%d", i);
                    std::string of = outfile;
                    of.insert(of.find_last_of(".dot")-3, buffer);
                    GraphExport::toDotFile(of, g, i);
                }
            }
        } catch(GraphIOException exc) {
            std::cout << exc.getCause() << std::endl;
            return -1;
        }
    }

    return 0;
}

