// KMB 2006 Jul 19

#include <limits.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/times.h>

#include "vn_graph.h"

//#define MALLOC_FREE_CHECK
#ifdef MALLOC_FREE_CHECK
  static size_t alloc=0;
  #define Malloc(sz) malloc((alloc+=(sz),sz))
  #define Calloc(n,sz) calloc(n,(alloc+=(n*(sz)),sz))
  #define Realloc(p,sz) realloc(p,(sz)) // FIXME
  #define Free(p,sz) free((alloc-=(sz),p))
  #define print_alloc(msg) fprintf(stderr,"%s alloc=%ld\n",msg,(long)alloc);
#else
  #define Malloc(sz) malloc(sz)
  #define Calloc(p,sz) calloc(p,sz)
  #define Realloc(p,sz) realloc(p,sz)
  #define Free(p,sz) free(p)
#endif

#include "trick.c"
#include "clique.c"
#include "rainbow.c"
#include "geng_reader.c"

static void shuffle(int p[], int n, int m);

//! make a new empty graph with n nodes
graph_t graph_new(unsigned int n) {
  int i,j;
  graph_t g=Malloc(sizeof(struct _graph));
  g->a=Malloc(n*sizeof(int*));
  g->v=Malloc(n*sizeof(int));
  g->c=Malloc(n*sizeof(int));
  g->l=Malloc(n*sizeof(int));
  for (i=0; i<n; i++) {
    g->c[i]=-1;
    g->l[i]=-1;
    g->v[i]=0;
    g->a[i]=Malloc(n*sizeof(int)); // no self-loops; one more for sentinel
    for (j=0; j<n; j++) g->a[i][j]=-1; // -1=top of list
  }
  g->nnodes=n;
  g->nedges=0;
  return g;
}

//! destroy a graph
void graph_clear(graph_t g) {
  int i,n=g->nnodes;
  for (i=0; i<n; i++) Free(g->a[i],n*sizeof(int));
  Free(g->l,n*sizeof(int));
  Free(g->c,n*sizeof(int));
  Free(g->v,n*sizeof(int));
  Free(g->a,n*sizeof(int*));
  Free(g,sizeof(struct _graph));
}

//! add a node to a graph
void graph_add_node(graph_t g) {
  unsigned int i,j,n=g->nnodes;
  for (i=0; i<n; i++) {
    g->a[i]=Realloc(g->a[i],(n+1)*sizeof(int));
    g->a[i][n]=-1;
  }
  n++;
  g->a=Realloc(g->a,n*sizeof(int*));
  g->a[n-1]=Malloc(n*sizeof(int));
  for (j=0; j<n; j++) g->a[n-1][j]=-1;
  g->v=Realloc(g->v,n*sizeof(int));
  g->c=Realloc(g->c,n*sizeof(int));
  g->l=Realloc(g->l,n*sizeof(int));
  for (i=0; i<n; i++) {
    g->v[i]=0;
    g->c[i]=-1;
    g->l[i]=-1;
  }
  g->nnodes=n;
} 

//! empty a graph
void graph_empty(graph_t g) {
  int i,j;
  for (i=0; i<g->nnodes; i++) {
    for (j=0; j<g->nnodes; j++) g->a[i][j]=-1;
  }
  g->nedges=0;
}

//! add undirected edge i--j to the graph, if it not already present
void graph_add_edge(graph_t g, node_t i, node_t j) {
  int k,l;
  k=-1;
  while ((l=g->a[i][++k])!=-1) if (l==j) return;
  g->a[i][k]=j;
  k=-1;
  while ((l=g->a[j][++k])!=-1) if (l==i) return;
  g->a[j][k]=i;
  g->nedges++;
}

//! degree of a node
int graph_node_degree(graph_t g, node_t i) {
  int k=-1;
  while (g->a[i][++k]!=-1);
  return k;
}

//! maximum degree of a node in g
int graph_max_degree(graph_t g) {
  int i,d,k=-1;
  for (i=0; i<g->nnodes; i++) if ((d=graph_node_degree(g,i))>k) k=d;
  return k;
}

//! mean degree
double graph_mean_degree(graph_t g) {
  return (2.0*g->nedges)/g->nnodes;
}

//! simple listing of graph
void graph_show(graph_t g) {
  int i,j;
  for (i=0; i<g->nnodes; i++) if (1||g->a[i][0]>=0) { // i has neighbours
    printf("%d: ",i);
    for (j=0; j<g->nnodes; j++) if (g->a[i][j]>=0) printf("%d ",g->a[i][j]);
    printf("\n");
  }
}

void graph_show_edge_list(graph_t g, int offset) {
  int i,j;
  printf("%d %d\n",g->nnodes,g->nedges);
  for (i=0; i<g->nnodes; i++) if (g->a[i][0]>=0) { // i has neighbours
    for (j=0; j<g->nnodes; j++) if (g->a[i][j]>=0 && g->a[i][j]>i) printf("%d %d\n",i+offset,g->a[i][j]+offset);
  }
}

void graph_dump_to_file(graph_t g, char* fn) {
  // my edge list format
  int i,j,k;
  FILE* f=fopen(fn,"w");
  fprintf(f,"n=%d m=%d.\n",g->nnodes,g->nedges);
  for (i=0; i<g->nnodes; i++) if (g->a[i][0]>=0) { // i has neighbours
    k=-1;
    while ((j=g->a[i][++k])>=0) if (j>i) fprintf(f,"%d %d\n",i,j);
  }
  fclose(f);
  printf("%s written\n",fn);
}

void graph_to_showg(graph_t g, int n, char* fn) {
  // "showg -e -l0" format
  int i,j,k;
  FILE* f=fopen(fn,"w");
  fprintf(f,"Graph %d, order %d.\n",n,g->nnodes);
  fprintf(f,"%d %d\n",n,g->nedges);
  for (i=0; i<g->nnodes; i++) if (g->a[i][0]>=0) { // i has neighbours
    k=-1;
    while ((j=g->a[i][++k])>=0) if (j>i) fprintf(f,"%d %d  ",i,j);
  }
  fprintf(f,"\n");
  fclose(f);
  printf("%s written:\n",fn);
}

void graph_to_dimacs(graph_t g, char* fn) {
  // ftp://dimacs.rutgers.edu/pub/challenge/graph/doc/ccformat.tex
  // 1-indexed
  int i,j,k;
  FILE* f=fopen(fn,"w");
  fprintf(f,"c written by graph_to_dimacs\n");
  fprintf(f,"p edge %d %d\n",g->nnodes,g->nedges);
  for (i=0; i<g->nnodes; i++) if (g->a[i][0]>=0) { // i has neighbours
    k=-1;
    while ((j=g->a[i][++k])>=0) if (j>i) fprintf(f,"e %d %d\n",i+1,j+1);
  }
  fclose(f);
  printf("%s written\n",fn);
}

void graph_to_theta(graph_t g, char* fn) {
  // 1-indexed, complemented, for DSDP Lovasz theta
  int i,j,k,n=g->nnodes;
  FILE* f=fopen(fn,"w");
  int u[g->nnodes];
  fprintf(f,"%d %d\n",n,n*(n-1)/2-g->nedges);
  for (i=0; i<n; i++) if (g->a[i][0]>=0) { // i has neighbours
    bzero(u,n*sizeof(int));
    k=-1;
    while ((j=g->a[i][++k])>=0) u[j]=1;
    for (j=i+1; j<n; j++) if (!u[j]) fprintf(f,"%d %d\n",i+1,j+1);
  }
  fclose(f);
  printf("%s written: now do \"theta %s\"\n",fn,fn);
}

void graph_make_dotfile(graph_t g, char* fn) {
  int i,j,k;
  FILE* f=fopen(fn,"w");
  fprintf(f,"graph %s {\n","graph_make_dotfile");
  fprintf(f,"  center=true; unflatten=true;\n ratio=1;\n /* mclimit=5;\n nslimit=500;\n nslimit1=500;\n remincross=true; */\n size=\"8,8\";\n node  [shape=\"circle\",size=\"0.1\"]\n node [color=green,fontcolor=red,fontsize=8]\n edge [color=blue];\n graph [concentrate=\"true\"];\n");
  for (i=0; i<g->nnodes; i++) if (g->a[i][0]>=0) { // i has neighbours
    k=-1;
    while ((j=g->a[i][++k])>=0) if (j>i) fprintf(f,"  %d--%d;\n",i,j);
  }
  fprintf(f,"}\n");
  fclose(f);
  printf("%s written: now do \"neato -Tps %s | gv -\"\n",fn,fn);
}

void rgb2hexstring(double r, double g, double b, char *s) {
  double gi=1.0/1.0; // inverse gamma
  r=floor(255.9999999*pow(r,gi));
  g=floor(255.9999999*pow(g,gi));
  b=floor(255.9999999*pow(b,gi));
  snprintf(s,7,"%02x%02x%02x",(int)r,(int)g,(int)b);
}

// http://www.graphviz.org/doc/info/colors.html
void graph_make_dotfile_colored(graph_t g, char* fn) {
  int i,j,k,m=-1;
  char s[7];
  double rd,gr,bl;
  FILE* f=fopen(fn,"w");
  fprintf(f,"graph %s {\n","graph_make_dotfile_colored");
  for (i=0; i<g->nnodes; i++) if (g->c[i]>m) m=g->c[i];
  fprintf(f,"  center=true; unflatten=true;\n ratio=1;\n /* mclimit=5;\n nslimit=500;\n nslimit1=500;\n remincross=true; */\n size=\"8,10\";\n node [shape=\"circle\",size=\"0.03\"]\n node [fontsize=10,style=filled]\n edge [style=\"setlinewidth(1)\",color=black];\n graph [concentrate=\"true\",label=\"%d colors\"];\n",m);
  for (i=0; i<g->nnodes; i++) {
    rainbow((g->c[i])/(double)m,1.0,1.0,&rd,&gr,&bl);
    rgb2hexstring(rd,gr,bl,s);
    //fprintf(stderr,"%d %d %f %s\n",i,g->c[i],(g->c[i])/(double)m,s);
    fprintf(f,"  %d [color=\"#%s\",fillcolor=\"#%s\"];\n",i,s,s);
  }
  for (i=0; i<g->nnodes; i++) if (g->a[i][0]>=0) { // i has neighbours
    k=-1;
    while ((j=g->a[i][++k])>=0) 
      if (j>i) fprintf(f,"  %d--%d;\n",i,j);
  }
  fprintf(f,"}\n");
  fclose(f);
  printf("%s written: now do \"neato -Tps %s | gv -\"\n",fn,fn);
}

void graph_dfs(graph_t g, node_t s, int l) {
  int i,j,m=-1,*gas=g->a[s];
  g->l[s]=l; // cluster number
  g->v[s]=1; // we have just visited node s
  // find biggest color used by neighbours
  i=0; while ((j=gas[i++])>=0) if ((j=color(g,j))>=0) m=j>m?j:m;
  if (m==-1)      //  no neighbours are yet colored
    g->c[s]=0;  
  else if (m==0)  // one neighbours colored 0
    g->c[s]=1;
  else { // find colors used by neighbours - they are in the range [0,m]
    int u[m+2]; bzero(u,(m+2)*sizeof(int)); // extra 0 element as sentinel
    i=0; while ((j=gas[i++])>=0) if ((j=color(g,j))>=0) u[j]=1;
    // find smallest color not used by neighbours
    i=0; while (u[i++]); color(g,s)=i-1;
  }
  // visit neighbours
  i=0; while ((j=gas[i++])>=0) if (!visited(g,j)) graph_dfs(g,j,l);
}

int graph_greedy_color(graph_t g, int perm[]) {
  // greedy coloring
  int i,pi,c=0;
  for (i=0; i<g->nnodes; i++) { g->v[i]=0; g->c[i]=-1; } // -1=not colored
  for (i=0; i<g->nnodes; i++) {
    if (perm) pi=perm[i]; else pi=i;
    if (!visited(g,pi)) graph_dfs(g,pi,c++);
  }
  c=-1;
  for (i=0; i<g->nnodes; i++) if (color(g,i)>c) c=color(g,i);
  return c+1; // colors used
}

int graph_sequential_color(graph_t g, int perm[], int ub) {
  // N[i][j] => node i has a neighbour colored j
  // colored in order given by perm
  // return -1 as soon as more colors than ub used
  int i,j=0,k,l,m=-1,pi,**N=Malloc(g->nnodes*sizeof(int*));
  int *Npi,*gapi;
  for (i=0; i<g->nnodes; i++) N[i]=Calloc(g->nnodes,sizeof(int));
  for (i=0; i<g->nnodes; i++) {
    if (perm) pi=perm[i]; else pi=i;
    Npi=N[pi]; gapi=g->a[pi];
    j=-1; while (Npi[++j]);
    k =0; while ((l=gapi[k++])>=0) N[l][j]=1;
    g->c[pi]=j;
    if (j+1>ub) { m=-2; goto quit; };
    m=j>m?j:m;
  }
  quit: for (i=0; i<g->nnodes; i++) Free(N[i],g->nnodes*sizeof(int));
  Free(N,g->nnodes*sizeof(int*));
  return m+1; // number of colors used
}

//! return an upper bound to chi
int graph_sequential_color_repeat(graph_t g, int n) {
  int k,perm[g->nnodes],j,chi_ub=INT_MAX;
  for (j=0; j<g->nnodes; j++) perm[j]=j;
  for (j=0; j<n; j++) {
    k=graph_sequential_color(g,perm,chi_ub);
    if (k>0 && k<chi_ub) chi_ub=k; // got a better bound
    shuffle(perm,g->nnodes,g->nnodes);
  }
  return chi_ub;
}

int graph_check_coloring(graph_t g) {
  int i,j,k,ci;
  for (i=0; i<g->nnodes; i++) {
    ci=g->c[i];
    j=0;
    while ((k=g->a[i][j++])>=0) // for each neighbour
      if (ci==g->c[k]) { // something wrong
          //for (j=0; j<g->nnodes; j++) printf("color(%d)=%d\n",j,g->c[j]);
        return 0;
      }
  }
  return 1;  // ok
}

void graph_gnp(graph_t g, double p) {
  int i,j;
  graph_empty(g);
  for (i=0; i<g->nnodes; i++) for (j=i+1; j<g->nnodes; j++)
    if (rand()<p*RAND_MAX) graph_add_edge(g,i,j);
}

void graph_grg(graph_t g, double r) {
  int i,j,n=g->nnodes;
  double r2=r*r,dx,dy,x[n],y[n];
  graph_empty(g);
  for (i=0; i<n; i++) {
    x[i]=rand()/(RAND_MAX+1.0);
    y[i]=rand()/(RAND_MAX+1.0);
  }
  for (i=0; i<n; i++)
    for (j=i+1; j<n; j++) {
      dx=x[i]-x[j];
      dy=y[i]-y[j];
      if (dx*dx+dy*dy<r2) graph_add_edge(g,i,j);
  }
}

typedef struct {double x,y; } xy;

static int cmp(const void *a, const void *b) {
  xy *aa=(xy*)a;
  xy *bb=(xy*)b;
  return aa->x<bb->x?-1:(aa->x>bb->x?1:0);
}

static double torus_distance2(double dx, double dy) {
  if (dx>0.5) dx=1.0-dx;
  if (dy>0.5) dy=1.0-dy;
  return dx*dx+dy*dy;
}

void graph_grg_torus(graph_t g, double r) {
  int i,j,n=g->nnodes;
  double r2=r*r,x[n],y[n];
  graph_empty(g);
  for (i=0; i<n; i++) {
    x[i]=rand()/(RAND_MAX+1.0);
    y[i]=rand()/(RAND_MAX+1.0);
  }
  for (i=0; i<n; i++)
    for (j=i+1; j<n; j++)
      if (torus_distance2(fabs(x[i]-x[j]),fabs(y[i]-y[j]))<r2) graph_add_edge(g,i,j);
}

void graph_grg_torus_sorted(graph_t g, double r) {
  int i,j,m=0,n=g->nnodes;
  double r2=r*r;
  xy z[n];
  for (i=0; i<n; i++) {
    z[i].x=rand()/(RAND_MAX+1.0);
    z[i].y=rand()/(RAND_MAX+1.0);
  }
  qsort(z,n,sizeof(xy),cmp); // sort on x
  graph_empty(g);
  for (i=0; i<n; i++) {
    for (j=i+1; j<n; j++)
      if (torus_distance2(fabs(z[i].x-z[j].x),fabs(z[i].y-z[j].y))<r2) graph_add_edge(g,i,j),m++;
  }
}

int graph_ncolors(graph_t g) {
  int i,mc=-1;
  for (i=0; i<g->nnodes; i++) if (g->c[i]>mc) mc=g->c[i];
  return mc+1;
}

int graph_nclusters(graph_t g) {
  int i,ml=-1;
  graph_greedy_color(g,NULL);
  for (i=0; i<g->nnodes; i++) if (g->l[i]>ml) ml=g->l[i];
  return ml+1;
}

int graph_connected(graph_t g) {
  return graph_nclusters(g)==1?1:0;
}

// histogram functions

histogram_t histogram_new(char* lbl) {
  int sz=1000;
  histogram_t h=Malloc(sizeof(struct _histogram));
  h->h=Calloc(sz,sizeof(int));
  h->m=Calloc(sz,sizeof(double));
  h->ss=Calloc(sz,sizeof(double));
  h->a=sz;
  h->n=0;
  h->lbl=lbl;
  return h;
}

void histogram_add(histogram_t h, int i) {
  double om,x;
  if (i>=h->a) {
    int j=h->a;
    h->a=i+10;
    h->h=Realloc(h->h,h->a*sizeof(int));
    for (; j<h->a; j++) h->h[j]=0;
  }
  h->n++;
  h->h[i]++;
  // get the mean proportion so far in bin i...
  x=h->h[i]/(double)h->n; // the value being included in the mean and ss
  om=h->m[i];
  h->m[i]+=(x-om)/h->h[i];
  h->ss[i]+=(x-h->m[i])*(x-om);
}

void histogram_clear(histogram_t h) {
  Free(h->h,0);
  Free(h->m,0);
  Free(h->ss,0);
  Free(h,0);
}

void histogram_write(FILE* f, histogram_t h) {
  int i;
  int i0,i1,imx=-1,mx=-1;
  double se,sum=0.0,ave=0.0,ss=0.0;
  i0=-1;   while (i0<h->a-1 && !h->h[++i0]);
  i1=h->a; while (i1>i0     && !h->h[--i1]);
  for (i=i0; i<=i1; i++) {
    if ((int)h->h[i]>mx) { mx=h->h[i]; imx=i; }
    sum+=h->h[i];
    ave+=h->h[i]*i;
  }
  if (sum) {
    ave/=sum;
    fprintf(f,"%s %ld items; ",h->lbl,(long)sum);
    fprintf(f,"mode=%d",imx);
    fprintf(f," mean=%.6f",ave);
    for (i=i0; i<=i1; i++) ss+=h->h[i]*(i-ave)*(i-ave);
    fprintf(f," var=%.4g",ss/(sum-1));
    fprintf(f," stdev=%.4g",sqrt(ss/(sum-1)));
    fprintf(f," stderr=%.4g\n",sqrt(ss/(sum-1))/sqrt(sum));
    for (i=i0; i<=i1; i++) {
      se=sqrt(h->ss[i]/(h->h[i]-1))/sqrt(h->h[i]); // standard error of the mean
      fprintf(f,"%2d %10d   %8.6f   %6.3f%% se=%.3g\n",
        i,h->h[i],h->h[i]/(double)sum,100*h->h[i]/(double)sum,se);
    }
  }
}

void histogram_show(histogram_t h) {
  histogram_write(stdout,h);
}

void histogram_show_old(histogram_t h) {
  int i;
  int i0,i1;
  double se,sum=0.0,ave=0.0,ss=0.0;
  i0=-1;     while (i0<=h->a && !h->h[++i0]);
  i1=h->a+1; while (i1>i0    && !h->h[--i1]);
  for (i=i0; i<=i1; i++) {
    sum+=h->h[i];
    ave+=h->h[i]*i;
  }
  if (sum) {
    ave/=sum;
    printf("Histogram %s: %ld items, ",h->lbl,(long)sum);
    printf("average=%.6f ",ave);
    for (i=i0; i<=i1; i++) ss+=h->h[i]*(i-ave)*(i-ave);
    printf(" variance=%.4g ",ss/(sum-1));
    printf(" stdev=%.4g\n",sqrt(ss/(sum-1)));
    for (i=i0; i<=i1; i++) {
      se=sqrt(h->ss[i]/(h->h[i]-1))/sqrt(h->h[i]); // standard error of the mean
      printf("%2d %10d   %8.6f   %6.3f%% se=%.3g\n",
        i,h->h[i],h->h[i]/(double)sum,100*h->h[i]/(double)sum,se);
    }
  }
}

double histogram_quantile(histogram_t h, double y) {
  double cum(double x) { // cumulative distribution from h
    int i=0;
    double s=0.0,x0,x1,y0,y1;
    x0=floor(x-0.5)+0.5;
    for (i=0; i<(x0); i++) s+=h->h[i];
    x1=x0+1.0;
    y0=s;
    y1=s+h->h[i];
    return ((y1-y0)/(x1-x0)*(x-x0)+y0)/h->n; // linear interpolation
  }
  // bisect
  double md,dn=0,up=h->n;
  md=0.5*(up+dn);
  while (fabs(up-dn)>1e-9) {
    if (cum(md)>y) up=md; else dn=md;
    md=0.5*(up+dn);
  }
  return md;
}

double histogram_mean(histogram_t h) {
  unsigned int i;
  double m=0.0;
  for (i=0; i<h->a; i++) m+=h->h[i]*i;
  return m/h->n;
}

unsigned int histogram_mode(histogram_t h) {
  unsigned int i,mx=0,imx=0;
  for (i=0; i<h->a; i++) if (h->h[i]>mx) { mx=h->h[i]; imx=i; }
  return imx;
}

int histogram_biggest_done(histogram_t h, double eps) {
  // true if standard error of the mean of the highest bin is<eps
  unsigned int i,mx=0,imx=0;
  for (i=0; i<h->a; i++) if (h->h[i]>mx) { mx=h->h[i]; imx=i; }
  return sqrt(h->ss[imx])<eps*mx;
}

int histogram_mean_done(histogram_t h, double eps) {
  // true if standard error of the mean is<eps*(1+mean)
  unsigned int i;
  double m=0.0,s=0.0,v;
  for (i=0; i<h->a; i++) {
    m+=h->h[i]*i;
    s+=h->h[i]*i*i;
  }
  //fprintf(stderr,"m=%g s=%g %g; ",m,s,m*m/h->n);
  v=(s-m*m/h->n)/(h->n-1); // variance
  m/=h->n;                 // mean
  s=sqrt(v);               // stdev
  //fprintf(stderr,"%g %g %g\n",v,s/sqrt(h->n),eps*(1.0+m));
  return s/sqrt(h->n)<eps*(1.0+m);
}

#define histogram_bin(h,i) (h->h[i])
#define histogram_rel_bin(h,i) (h->h[i]/(double)h->n)

int graph_chromatic_number(graph_t g, clock_t timeout) {
  int c,i,j,k,l,place,*valid,*clique;
  #ifdef MALLOC_FREE_CHECK
    alloc=0;
    print_alloc("graph_chromatic_number start")
  #endif
  for (i=0; i<g->nnodes; i++) g->c[i]=-1; // in case already colored
  if (g->nedges==0) { // empty graph
    for (i=0; i<g->nnodes; i++) g->c[i]=0;
    return 1;
  }
  if (g->nedges==1) {
    for (i=0; i<g->nnodes; i++) g->c[i]=0;
    for (i=0; i<g->nnodes; i++) { // find the edge
      k=-1; while ((j=g->a[i][++k])==-1);
      g->c[j]=1;
      return 2;
    }
  }
  if (2*g->nedges==g->nnodes*(g->nnodes-1)) { // complete graph
    for (i=0; i<g->nnodes; i++) g->c[i]=i;
    return g->nnodes;
  }
  cg.adj=Malloc(g->nnodes*sizeof(int*));
  for (i=0; i<g->nnodes; i++) {
    cg.adj[i]=Calloc(g->nnodes,sizeof(int));
    k=0; while ((j=g->a[i][k++])>=0) cg.adj[i][j]=1;
  }
  cg.g=g;
  cg.num_node  =g->nnodes;
  valid        =Malloc(cg.num_node*sizeof(int));
  clique       =Calloc(cg.num_node,sizeof(int));
  cg.ColorClass=Malloc(cg.num_node*sizeof(int));
  cg.Order     =Malloc(cg.num_node*sizeof(int));
  cg.Handled   =Malloc(cg.num_node*sizeof(int));
  cg.ColorCount=Malloc(cg.num_node*sizeof(int));
  cg.ColorAdj  =Malloc(cg.num_node*sizeof(int*));
  for (i=0; i<cg.num_node; i++)
    cg.ColorAdj[i]=Malloc((cg.num_node+1)*sizeof(int));
  cg.prob_count=0;
  for (i=0; i<cg.num_node; i++) {
    for (j=0; j<=cg.num_node; j++) cg.ColorAdj[i][j]=0;
    for (j=0; j<cg.num_node; j++) if (cg.adj[i][j]) cg.ColorAdj[i][0]++;
    cg.ColorCount[i]=0;
    cg.Handled[i]=0;
    valid[i]=1;
  }
  cg.BestColoring=cg.num_node + 1;
  cg.best_clique=0;
  cg.lb=max_w_clique(valid,clique,0,cg.num_node);
  cg.num_prob=0;
  cg.max_prob=10000;
  // timing...
  cg.tps=sysconf(_SC_CLK_TCK); // clock ticks per second 
  times(&cg.buffer);
  cg.start_time=cg.buffer.tms_utime;
  cg.timeout=timeout;
  if (1) {
    place=0;
    for (i=0; i<cg.num_node; i++) if (clique[i]) {
      cg.Order[place]=i;
      cg.Handled[i]=1;
      place++;
      AssignColor(i,place);
      for (j=0; j<cg.num_node; j++)
        if (i!=j && clique[j] && !cg.adj[i][j]) {
          fprintf(stderr,"graph_chromatic_number internal error: result is not a clique!\n");
          graph_show(g);
          for (l=0; l<cg.num_node; l++) printf("clique[%d]=%d\n",l,clique[l]);
          for (l=0; l<cg.num_node; l++) {
            for (k=0; k<cg.num_node; k++) printf("%d ",cg.adj[l][k]);
            printf("\n");
          }
          exit(2);
        }
    }
    c=Color(place,place);
  }
  for (i=0; i<cg.num_node; i++) {
    Free(cg.ColorAdj[i],(cg.num_node+1)*sizeof(int));
    Free(cg.adj[i],g->nnodes*sizeof(int));
  }
  Free(cg.adj,g->nnodes*sizeof(int*));
  Free(cg.ColorAdj,cg.num_node*sizeof(int*));
  Free(cg.ColorClass,cg.num_node*sizeof(int));
  Free(cg.Order,cg.num_node*sizeof(int));
  Free(cg.Handled,cg.num_node*sizeof(int));
  Free(cg.ColorCount,cg.num_node*sizeof(int));
  Free(valid,cg.num_node*sizeof(int));
  Free(clique,cg.num_node*sizeof(int));
  #ifdef MALLOC_FREE_CHECK
    print_alloc("graph_chromatic_number end")
  #endif
  return c;
}

int graph_clique_number(graph_t g) {
  int i,j,k;
  int u[g->nnodes];
  clqdata.matrix_size=g->nnodes;
  clqdata.deg=Calloc(clqdata.matrix_size,sizeof(int));
  clqdata.IS=Calloc(clqdata.matrix_size,sizeof(int));
  clqdata.adj_list=Malloc(clqdata.matrix_size*sizeof(int*));
  clqdata.bucket=Malloc(clqdata.matrix_size*sizeof(int*));
  for (i=0; i<clqdata.matrix_size; i++) {
    clqdata.adj_list[i]=Malloc(clqdata.matrix_size*sizeof(int));
    clqdata.bucket[i]=Calloc(clqdata.matrix_size,sizeof(int));
    for (j=0; j<clqdata.matrix_size; j++) clqdata.adj_list[i][j]=1;
  }
  for (i=0; i<g->nnodes; i++) {
    bzero(u,clqdata.matrix_size*sizeof(int));
    k=-1;
    while ((j=g->a[i][++k])>=0) u[j]=1;
    for (j=0; j<clqdata.matrix_size; j++) if (!u[j] && i!=j) { // i!-j
      clqdata.adj_list[i][clqdata.deg[i]++]=j; // complement of g
    }
  }
  clqdata.maxsz=-1;
  clqdata.mode=0;
  clqdata.verbose=0;
  backtrack(0);
  //graph_clqdata_al_show();
  for (i=0; i<clqdata.matrix_size; i++) {
    Free(clqdata.adj_list[i],0);
    Free(clqdata.bucket[i],0);
  }
  Free(clqdata.deg,0);
  Free(clqdata.IS,0);
  Free(clqdata.adj_list,0);
  Free(clqdata.bucket,0);
  return clqdata.maxsz;
}

//! read and process "showg -l0 -e" output
histogram_t graph_showg_reader(FILE* f, int op(graph_t)) { 
  int i,j,k,verbose=0,num_node,num_graph,m;
  graph_t g;
  histogram_t h=histogram_new("graph_showg_reader");
  while (!feof(f)) { // loop over graphs in "showg -e -l0" format
    i=scanf(" Graph %d, order %d.",&num_graph,&num_node);
    if (i<=0) break; // done
    if (verbose) printf("Graph %d with %d nodes, ",num_graph,num_node);
    scanf("%d %d",&num_node,&m);
    if (verbose) printf("%d edges\n",m);
    g=graph_new(num_node);
    for (k=0; k<m; k++) {
      scanf("%d %d",&i,&j);
      graph_add_edge(g,i,j);
    }
    histogram_add(h,op(g));
    graph_clear(g);
  }
  return h;
}

//! read and process "geng" output
histogram_t graph_geng_reader(FILE* f, int op(graph_t), char* lbl) {
  char *s;
  graph_t g;
  histogram_t h=histogram_new(lbl);
  while ((s=geng_getline(stdin))) {
    g=geng_stringtograph(s);
    //graph_show(g);
    histogram_add(h,op(g));
    if (s) free(s);
    graph_clear(g);
  }
  return h;
}

unsigned int xcng(void) {
  static unsigned int x=12345689,y=521289,z,w;
  z=x+y; x=y; w=(z<<8); y=w+17*(z>>24);
  if (z<x) y+=4352; if (y<w) y+=17; return y;
}

double uniform01(void) {
  return 2.3283064365386962891e-10*xcng();
}

unsigned int randint(int upto) {
  return (unsigned int)upto*((double)xcng()/(double)UINT_MAX);
  //return (int)upto*uniform01();
}

static void shuffle(int p[], int n, int m) {
  // uses a different RNG so we get the same sequence of random graphs
  int i,j,k,l,p0=p[0];
  for (k=0; k<m; k++) {
    i=randint(n);
    while ((j=randint(n))==i);
    l=p[i]; p[i]=p[j]; p[j]=l;
  }
  if (p0==p[0]) { // make sure 0 is not fixed
    while ((j=randint(n))==0);
    l=p[0]; p[0]=p[j]; p[0]=l;
  }
}
